# Copyright 2010 University of Chicago
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'trhowe'


RELATIONSHIP_INDEX = '__relationship__index__'
OUTBOUND_RELATIONSHIP_CF = 'outbound__%s' % RELATIONSHIP_INDEX
INBOUND_RELATIONSHIP_CF = 'inbound__%s' % RELATIONSHIP_INDEX
RELATIONSHIP_KEY_PATTERN = '%s__%s'
ENDPOINT_NAME_TEMPLATE = '%s__%s'
